import abc
import contextlib
import enum
from typing import Dict, List, Optional, Set, Tuple, Union

from ccimport import compat

from pccm.core import Argument, FunctionCode
from pccm.utils import UniqueNamePool


class PointerType(enum.Enum):
    Global = "Global"
    Smem = "Smem"
    ExternalRegister = "ExternalRegister"


class RegDType(enum.Enum):
    F16 = "f16,16"
    F16X2 = "f16x2,32"
    F32 = "f32,32"
    F64 = "f64,64"
    S8 = "s8,8"
    S16 = "s16,16"
    S32 = "s32,32"
    S64 = "s64,64"
    U8 = "u8,8"
    U16 = "u16,16"
    U32 = "u32,32"
    U64 = "u64,64"
    B8 = "b8,8"
    B16 = "b16,16"
    B32 = "b32,32"
    B64 = "b64,64"
    BF16 = "bf16,16"
    TF32 = "tf32,32"
    PRED = "pred,1"


class OperandBase:
    def __init__(self, name: str):
        self.is_input = False
        self.is_output = False
        self.index = -1
        self.offset = 0
        self.name = name

    def __hash__(self):
        return hash(self.name)


class PointerBase(OperandBase):
    def __init__(self, name: str, ptr_type: PointerType):
        super().__init__(name)
        self.ptr_type = ptr_type

    def get_inline_asm_dtype(self):
        # all pointers are 64bit EXCEPT smem
        return "l" if self.ptr_type == PointerType.Global else "r"


class Pointer(PointerBase):
    def __init__(self, name: str, ptr_type: PointerType):
        super().__init__(name, ptr_type)

    def copy(self):
        res = Pointer(self.name, self.ptr_type)
        res.is_input = self.is_input
        res.is_output = self.is_output
        res.index = self.index
        res.offset = self.offset
        return res

    def __add__(self, offset: int):
        assert self.ptr_type != PointerType.ExternalRegister
        res = self.copy()
        res.offset += offset
        return res


class RegisterBase(OperandBase):
    def __init__(self, name: str, dtype: RegDType, count: int = 1):
        super().__init__(name)
        self.dtype = dtype
        dtype_value: List[str] = dtype.value.split(",")
        self.dtype_str = dtype_value[0]
        self.bitsize = int(dtype_value[1])
        self.count = count
        self.is_addr_register = False
        assert count >= 1 and count <= 4

    def __hash__(self):
        return hash(self.name)


class Register(RegisterBase):
    def decl_stmt(self) -> str:
        if self.count > 1:
            return f".reg .v{self.count} .{self.dtype_str} {self.name};"
        else:
            return f".reg .{self.dtype_str} {self.name};"

    def copy(self):
        res = Register(self.name, self.dtype, self.count)
        res.is_input = self.is_input
        res.is_output = self.is_output
        res.index = self.index
        res.offset = self.offset

        res.is_addr_register = self.is_addr_register
        # other attrs are immutable.
        return res

    def __add__(self, offset: int):
        res = self.copy()
        res.offset += offset
        return res

    def as_addr_reg(self):
        res = self.copy()
        res.is_addr_register = True
        return res

    def as_output_reg(self):
        res = self.copy()
        res.is_output = True
        return res


class ExternalRegisterPointer(PointerBase):
    kAllowedExtDTypesToLetter = {
        RegDType.U16: "h",
        RegDType.U32: "r",
        RegDType.S32: "r",
        RegDType.B32: "r",
        RegDType.U64: "l",
        RegDType.F32: "f",
        RegDType.F64: "d",
    }

    def __init__(self, name: str, dtype: RegDType):
        assert dtype in self.kAllowedExtDTypesToLetter
        super().__init__(name, PointerType.ExternalRegister)
        self.dtype = dtype
        dtype_value: List[str] = dtype.value.split(",")
        self.dtype_str = dtype_value[0]
        self.bitsize = int(dtype_value[1])
        self._inline_asm_constraint_letter = self.kAllowedExtDTypesToLetter[
            self.dtype]

    def get_inline_asm_dtype(self):
        return self._inline_asm_constraint_letter

    def __getitem__(self, offset: int):
        return ExternalRegister(self.name + "[{}]".format(offset), self.dtype)

    def unpack(self, start: int, stop: Optional[int] = None):
        if stop is None:
            stop = start
            start = 0
            assert stop > 0
        return [self[i] for i in range(start, stop)]


class ExternalRegister(RegisterBase):
    def __init__(self, name: str, dtype: RegDType):
        super().__init__(name, dtype)
        self._inline_asm_constraint_letter = ExternalRegisterPointer.kAllowedExtDTypesToLetter[
            self.dtype]

    def get_inline_asm_dtype(self):
        return self._inline_asm_constraint_letter

    def copy(self):
        res = ExternalRegister(self.name, self.dtype)
        res.is_input = self.is_input
        res.is_output = self.is_output
        res.index = self.index
        res.offset = self.offset

        res.is_addr_register = self.is_addr_register
        # other attrs are immutable.
        return res

    def __add__(self, offset: int):
        res = self.copy()
        res.offset += offset
        return res

    def as_addr_reg(self):
        res = self.copy()
        res.is_addr_register = True
        return res

    def as_output_reg(self):
        res = self.copy()
        res.is_output = True
        return res


PTX_PARAMS_TYPES_BASE = Union[Pointer, Register, ExternalRegister, str, int]

PTX_PARAMS_TYPES = Union[PTX_PARAMS_TYPES_BASE, List[PTX_PARAMS_TYPES_BASE]]

PTX_OPERAND_TYPES = Union[Pointer, ExternalRegister]
CONSTANT_TYPES = Union[str, int]
REG_INPUT_OPERAND_TYPES = Union[Register, ExternalRegister, str, int]
REG_OPERAND_TYPES = Union[Register, ExternalRegister]


class AsmStmt(abc.ABC):
    @abc.abstractmethod
    def get_stmt_str(self) -> str:
        return ""

    @abc.abstractmethod
    def get_operands(self) -> List[PTX_OPERAND_TYPES]:
        """childs must determine pointer i/o attr
        before call this method.
        """
        return []

    @abc.abstractmethod
    def get_registers(self) -> List[Register]:
        return []


class RegDeclStmt(AsmStmt):
    def __init__(self, name: str, dtype: RegDType, count: int = 1):
        self.reg = Register(name, dtype, count)

    def get_stmt_str(self) -> str:
        return ""

    def get_operands(self) -> List[PTX_OPERAND_TYPES]:
        """childs must determine pointer i/o attr
        before call this method.
        """
        return []

    def get_registers(self) -> List[Register]:
        return [self.reg]


class Generic(AsmStmt):
    def __init__(self,
                 stmt: str,
                 params: List[PTX_PARAMS_TYPES],
                 pred_reg: str = "",
                 has_output: bool = False):
        self.stmt = stmt
        self.params = params
        self.operands: List[PTX_OPERAND_TYPES] = []
        self.registers: List[Register] = []

        self.pred_reg = pred_reg
        self.has_output = has_output
        for p in params:
            if isinstance(p, list):
                for pp in p:
                    assert not isinstance(pp, ExternalRegisterPointer)
                    if isinstance(pp, (ExternalRegister, Pointer)):
                        self.operands.append(pp)
                    elif isinstance(pp, Register):
                        self.registers.append(pp)
            else:
                assert not isinstance(p, ExternalRegisterPointer)
                if isinstance(p, (ExternalRegister, Pointer)):
                    self.operands.append(p)
                elif isinstance(p, Register):
                    self.registers.append(p)

    def get_stmt_str(self) -> str:
        op_strs: List[str] = []
        name_to_op_str: Dict[str, str] = {}
        for op in self.params:
            if not isinstance(op, list):
                ops = [op]
            else:
                ops = op
            for op in ops:
                if isinstance(op, Pointer):
                    # addr
                    if op.offset > 0:
                        op_str = "[%{}+{}]".format(op.index, op.offset)
                    else:
                        op_str = "[%{}]".format(op.index)
                elif isinstance(op, ExternalRegister):
                    if op.is_addr_register:
                        if op.offset > 0:
                            op_str = "[%{}+{}]".format(op.index, op.offset)
                        else:
                            op_str = "[%{}]".format(op.index)
                    else:
                        op_str = "%{}".format(op.index)
                elif isinstance(op, Register):
                    if op.is_addr_register:
                        if op.offset > 0:
                            op_str = "[{}+{}]".format(op.name, op.offset)
                        else:
                            op_str = "[{}]".format(op.name)
                    else:
                        op_str = "{}".format(op.name)
                else:
                    op_str = "{}".format(op)
                if not isinstance(op, (str, int)):
                    name_to_op_str[op.name] = op_str

        for i, op in enumerate(self.params):
            if isinstance(op, list):
                if len(op) == 1:
                    if not isinstance(op[0], (str, int)):
                        op_str_with_packed = name_to_op_str[op[0].name]
                    else:
                        op_str_with_packed = str(op[0])
                else:
                    # op must be list of registers
                    # TODO check this.
                    # packed param
                    op_strs_regs: List[str] = []
                    for opp in op:
                        assert isinstance(
                            opp,
                            (Register,
                             ExternalRegister)) and not opp.is_addr_register
                        op_strs_regs.append(name_to_op_str[opp.name])
                    # if i == 0 and self.has_output:
                    #     op_str_with_packed = "|".join(op_strs_regs)
                    # else:
                    op_str_with_packed = "{{{}}}".format(
                        ", ".join(op_strs_regs))
            else:
                if not isinstance(op, (str, int)):
                    op_str_with_packed = name_to_op_str[op.name]
                else:
                    op_str_with_packed = str(op)
            op_strs.append(op_str_with_packed)
        if self.pred_reg:
            return "@{} {} {};".format(self.pred_reg, self.stmt,
                                       ",".join(op_strs))
        else:
            return "{} {};".format(self.stmt, ",".join(op_strs))

    def get_operands(self) -> List[PTX_OPERAND_TYPES]:
        """childs must determine pointer i/o attr
        before call this method.
        """
        return self.operands

    def get_registers(self) -> List[Register]:
        return self.registers


class CacheOpLd(enum.Enum):
    # https://docs.nvidia.com/cuda/parallel-thread-execution/index.html#cache-operators
    AllLevel = "ca"
    L2AndBelow = "cg"
    Stream = "cs"
    LastUse = "lu"
    NoCache = "cv"


class CacheOpSt(enum.Enum):
    AllLevel = "wb"
    L2AndBelow = "cg"
    Stream = "cs"
    WriteThrough = "wt"


class PTXContext:
    """
    asm_stmt, operands (pointers and registers)
    for stmt in asm_stmts:
        ctx.query_decl("")
        ctx.assign_io(4) # io idx += 4
    https://docs.nvidia.com/cuda/parallel-thread-execution/index.html#comparison-and-selection-instructions-setp

    TODO handle addr from 64bit register
    """
    def __init__(self):
        self._asm_stmts: List[AsmStmt] = []
        self._pre_asm_stmts: List[str] = []
        self._cur_if_pred = ""

    @contextlib.contextmanager
    def cond(self, pred_reg: Register, not_pred: bool = False):
        assert not self._cur_if_pred
        name = pred_reg.name
        if not_pred:
            name = "!" + pred_reg.name
        self._cur_if_pred = name
        yield pred_reg
        self._cur_if_pred = ""

    @contextlib.contextmanager
    def pred_if(self, name: str, cmp: str, a: REG_INPUT_OPERAND_TYPES,
                b: REG_INPUT_OPERAND_TYPES):
        assert not self._cur_if_pred
        pred_reg = Register(name, RegDType.PRED)
        self.setp(cmp, pred_reg, a, b)
        self._cur_if_pred = name
        yield pred_reg
        self._cur_if_pred = ""

    @staticmethod
    def reg(name: str, dtype: RegDType, count: int = 1):
        return Register(name, dtype, count)

    @staticmethod
    def ext_reg(name: str, dtype: RegDType):
        return ExternalRegister(name, dtype)

    @staticmethod
    def global_ptr(name: str):
        return Pointer(name, PointerType.Global)

    @staticmethod
    def smem_ptr(name: str):
        return Pointer(name, PointerType.Smem)

    @staticmethod
    def reg_ptr(name: str, dtype: RegDType):
        return ExternalRegisterPointer(name, dtype)

    def generic(self,
                stmt: str,
                params: List[PTX_PARAMS_TYPES],
                has_output: bool = False):

        self._asm_stmts.append(
            Generic(stmt, params, self._cur_if_pred, has_output))

    def _ld_or_st(self, is_ld: bool, addr: Pointer,
                  regs: Union[List[REG_OPERAND_TYPES], REG_OPERAND_TYPES],
                  cache_op: Optional[Union[CacheOpLd, CacheOpSt]],
                level: str = "",
                prefetch_size: int = -1):
        if level:
            assert level == "L2"
            assert prefetch_size in [64, 128, 256]
        if not addr.is_input:
            addr.is_input = is_ld
        if is_ld:
            if cache_op is not None:
                assert isinstance(cache_op, CacheOpLd)
        else:
            if cache_op is not None:
                assert isinstance(cache_op, CacheOpSt)

        if not isinstance(regs, list):
            regs = [regs]
        assert len(regs) in [1, 2, 4], "vec load only support 124"
        for reg in regs:
            if not reg.is_output:
                reg.is_output = is_ld
        ld_or_st = "ld" if is_ld else "st"
        if addr.ptr_type == PointerType.Global:
            stmt = "{}.global".format(ld_or_st)
        else:
            stmt = "{}.shared".format(ld_or_st)
        if level:
            stmt += f".{level}::{prefetch_size}B"
        if len(regs) > 1:
            stmt += ".v{}".format(len(regs))
        if cache_op is not None:
            stmt += ".{}".format(cache_op.value)
        stmt += ".{}".format(regs[0].dtype_str)
        params: List[PTX_PARAMS_TYPES] = []
        if is_ld:
            params.append(list(regs))
            params.append(addr)
        else:
            params.append(addr)
            params.append(list(regs))
        self.generic(stmt, params, True)

    def setp(self, cmp_op: str, dst: REG_OPERAND_TYPES,
             a: REG_INPUT_OPERAND_TYPES, b: REG_INPUT_OPERAND_TYPES):
        dst.is_output = True
        dtype = ""
        if not isinstance(a, (str, int)):
            a.is_input = True
            dtype = a.dtype_str
        if not isinstance(b, (str, int)):
            b.is_input = True
            dtype = b.dtype_str
        assert dtype, "a or b must at least reg"
        stmt = "setp.{}.{}".format(cmp_op, dtype)
        self.generic(stmt, [dst, a, b], True)

    def ld(self,
           addr: Pointer,
           regs: Union[List[REG_OPERAND_TYPES], REG_OPERAND_TYPES],
           cache_op: Optional[CacheOpLd] = None,
           level: str = "",
           prefetch_size: int = -1):
        return self._ld_or_st(True, addr, regs, cache_op, level, prefetch_size)

    def st(self,
           addr: Pointer,
           regs: Union[List[REG_OPERAND_TYPES], REG_OPERAND_TYPES],
           cache_op: Optional[CacheOpSt] = None):
        return self._ld_or_st(False, addr, regs, cache_op)

    def mov(self, dst: REG_OPERAND_TYPES,
            srcs: Union[List[REG_INPUT_OPERAND_TYPES],
                        REG_INPUT_OPERAND_TYPES]):
        if not isinstance(srcs, list):
            srcs_check = [srcs]
        else:
            srcs_check = srcs
        for src in srcs_check:
            if not isinstance(src, (str, int)):
                src.is_input = True
        dst.is_output = True

        assert len(srcs_check) in [1, 2, 4], "vec load only support 124"
        params: List[PTX_PARAMS_TYPES] = []
        params.append(dst)
        params.append(srcs)
        self._asm_stmts.append(
            Generic("mov.{}".format(dst.dtype_str), params, self._cur_if_pred))

    def _get_str(self) -> str:
        ptr_idx = 0
        stmt_strs: List[str] = []
        name_to_operands: Dict[str, List[PTX_OPERAND_TYPES]] = {}
        all_outputs: List[PTX_OPERAND_TYPES] = []
        all_inputs: List[PTX_OPERAND_TYPES] = []
        # we need to group operands by name to
        # set index correctly.
        for stmt in self._asm_stmts:
            operands = stmt.get_operands()
            for op in operands:
                # print(op.is_input, op.is_output, op, op.name)
                if op.name not in name_to_operands:
                    name_to_operands[op.name] = []
                name_to_operands[op.name].append(op)
        for _, vs in name_to_operands.items():
            # reduce op setting with same name
            is_out = False
            is_inp = False
            leader = vs[0]
            for v in vs:
                if v.is_output:
                    is_out = True
                if v.is_input:
                    is_inp = True
            leader.is_output = is_out
            leader.is_input = is_inp
            if not leader.is_output:
                all_inputs.append(leader)
            else:
                all_outputs.append(leader)

        for op in all_outputs:
            for op_same_name in name_to_operands[op.name]:
                op_same_name.index = ptr_idx
            ptr_idx += 1
        for op in all_inputs:
            for op_same_name in name_to_operands[op.name]:
                op_same_name.index = ptr_idx
            ptr_idx += 1
        # after set index of ptrs, we can get real asm stmt.
        all_regs: Dict[str, Register] = {}
        for stmt in self._asm_stmts:
            regs = stmt.get_registers()
            for r in regs:
                if r.name not in all_regs:
                    all_regs[r.name] = r
                    stmt_strs.append(r.decl_stmt())
            stmt_str = stmt.get_stmt_str()
            if stmt_str:
                stmt_strs.append(stmt_str)
        asm_output_strs: List[str] = []
        asm_input_strs: List[str] = []
        # TODO check all ops with same name, they must have same type.
        for op in all_outputs:
            op_has_input = False
            for op_same_name in name_to_operands[op.name]:
                if op_same_name.is_input:
                    op_has_input = True
                    break
            if op_has_input:
                constraint_letter = "+"
            else:
                constraint_letter = "="
            asm_output_strs.append("\"{}{}\"({})".format(
                constraint_letter, op.get_inline_asm_dtype(), op.name))
        for op in all_inputs:
            asm_input_strs.append("\"{}\"({})".format(
                op.get_inline_asm_dtype(), op.name))
        outputs_str = ", ".join(asm_output_strs)
        inputs_str = ", ".join(asm_input_strs)
        stmt_prefix = "            "
        asm_stmts_str = "\n".join(
            [stmt_prefix + "\"  {}\\n\"".format(s) for s in stmt_strs])
        if len(asm_output_strs) == 0:
            if len(asm_input_strs) == 0 and len(stmt_strs) == 0:
                return ""
            return """
            asm volatile (
            "{{\\n"
{asm_stmts_str}
            "}}\\n"
            :: {inputs}
            );
            """.format(asm_stmts_str=asm_stmts_str, inputs=inputs_str)

        return """
        asm volatile (
            "{{\\n"
{asm_stmts_str}
            "}}\\n"
            : {outputs}
            : {inputs}
        );
        """.format(asm_stmts_str=asm_stmts_str,
                   outputs=outputs_str,
                   inputs=inputs_str)


class PTXCode(FunctionCode):
    """TODO
    A = code.global_ptr("", length=...)
    B = code.smem_ptr("", length=...)
    C = code.register_ptr("", length=...)
    with code.asm_if("var_name"):
        a, b, c = code.decl_reg("a, b, c")
        code.asm_assign(A + 8, B + 16)
        code.asm_assign(a, 0)

    """
    def __init__(self,
                 code: str = "",
                 arguments: Optional[List[Argument]] = None,
                 return_type: str = "void",
                 ctor_inits: Optional[List[Tuple[str, str]]] = None):
        super().__init__(code, arguments, return_type, ctor_inits)
        self._asm_mode = False

    @contextlib.contextmanager
    def asm_block(self):
        assert not self._asm_mode, "you can't enter asm block twice or more."
        ctx = PTXContext()
        self._asm_mode = True
        yield ctx
        self._asm_mode = False
        res = ctx._get_str()
        # print(res)
        self.raw(res)
        # raise NotImplementedError


if __name__ == "__main__":
    ctx = PTXContext()
    # ptr_g = ctx.global_ptr("A") + 4
    # ptr_er2 = ctx.reg_ptr("fragA", RegDType.U64)

    # ptr_er0 = ctx.reg_ptr("frag", RegDType.U32)
    # ctx.ld(ptr_g, [ptr_er0[0], ptr_er0[1]])
    # ctx.st(ptr_g, [ptr_er0[0], ptr_er0[1]])

    # ctx.mov(ptr_er0[0], ptr_er0[1])
    # ctx.mov(ptr_er2[0], ptr_er0.unpack(4))
    # with ctx.pred_if("p", "ne", ptr_er0[0], 0):
    #     ctx.mov(ptr_er0[0], ptr_er0[1])
    # ctx.mov(ptr_er0[0], ptr_er0[1])

    ptr_addr = ctx.global_ptr("addr")
    frag = ctx.reg_ptr("frag", RegDType.U32)

    pred = ctx.ext_reg("(int)pred", RegDType.U32)
    for i in range(8):
        ctx.mov(frag[i], 0)
    with ctx.pred_if("p", "ne", pred, 0):
        frag_unpack = frag.unpack(8)
        ctx.ld(ptr_addr, frag_unpack[:4])
        ctx.ld(ptr_addr + 8 * 4, frag_unpack[4:])

    print(ctx._get_str())



from pathlib import Path 

PCCM_FUNC_META_KEY = "__pccm_func_meta"
PCCM_CLASS_META_KEY = "__pccm_class_meta"

PCCM_MAGIC_STRING = "// THIS FILE IS AUTO-GENERATED BY PCCM, DON'T MODIFY THIS LINE!!! md5: {}"

PCCM_INIT_DECORATOR_KEY = "__pccm_this_type"

PCCM_INLINE_LIBRARY_PATH = Path.home() / ".pccm" / "inline"

PCCM_DISABLE_CODE_CHANGE = "PCCM_DISABLE_CODE_CHANGE"


# Copyright 2021 Yan Yan
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
# build a class
pccm pkg.mod1.mod2.class0 -o lib_name --build_dir ...
# build a param class
pccm "pkg.mod1.pclass[a=5, b=3, x=y]" -o lib_name --build_dir ...
# generate code and ninja build file
pccm-gen pkg.mod1.mod2.class0 xxx_folder 
pccm-gen "pkg.mod1.pclass[a=5, b=3, x=y]" xxx_folder 
# generate code and cmake files
pccm-gen-cmake pkg.mod1.mod2.class0 xxx_folder 
"""

import sys 
import importlib

def import_name(name, package=None):
    module = importlib.import_module(name, package)
    return module


def main_gen():
    pass

if __name__ == "__main__":
    print(sys.argv)


"""
we need simple structured code repr to print beautiful code.
class Block:
    prefix: str
    body: List[str|Block]
    suffix: str
"""
from typing import List, Optional, Union


class Block:
    def __init__(self,
                 prefix: str,
                 body: Optional[List[Union["Block", str]]] = None,
                 suffix: str = "",
                 indent: Optional[int] = None):
        self.prefix = prefix
        if body is None:
            body = []
        self.body = body
        self.suffix = suffix
        self.indent = indent

    def __repr__(self) -> str:
        return "\n".join(generate_code(self, 0, 2))


def generate_code(block: Union[Block, str], start_col_offset: int,
                  indent: int):
    col_str = " " * start_col_offset
    if isinstance(block, str):
        block_lines = block.split("\n")
        return [col_str + l for l in block_lines]
    res = []  # type: List[str]
    prefix = block.prefix
    next_indent = indent
    if block.indent is not None:
        next_indent = block.indent
    if prefix:
        prefix_lines = prefix.split("\n")
        res.extend([col_str + l for l in prefix_lines])
    for child in block.body:
        res.extend(generate_code(child, start_col_offset + next_indent,
                                 indent))
    if block.suffix:
        suffix_lines = block.suffix.split("\n")
        res.extend([col_str + l for l in suffix_lines])
    return res


def generate_code_list(blocklist: List[Union[Block, str]],
                       start_col_offset: int, indent: int):
    res = []  # type: List[str]
    for b in blocklist:
        res.extend(generate_code(b, start_col_offset, indent))
    return res



from pccm.core import Class


class CUDADriver(Class):
    def __init__(self):
        self.add_include("cuda_runtime_api.h")
        self.add_include("cuda.h")


class CUDADevice(Class):
    def __init__(self):
        self.add_include("cuda_fp16.h")


class CUBLAS(Class):
    def __init__(self):
        self.add_include("cublas.h")
        self.add_include("cublaslt.h")



import ast
import inspect
import json
import os
import threading
import time
from collections.abc import Mapping
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Type, Union

import pccm
import portalocker
from ccimport import loader, source_iter
from pccm.builder import build_pybind
from pccm.constants import PCCM_INLINE_LIBRARY_PATH
from pccm.core import (Class, ConstructorMeta, DestructorMeta,
                       ExternalFunctionMeta, FunctionCode, FunctionDecl,
                       MemberFunctionMeta, ParameterizedClass,
                       StaticMemberFunctionMeta, get_class_meta)
from pccm.core.parsers import BaseType, CppType, DeclSpec, QualifiedId
from pccm.middlewares.pybind import Pybind11MethodMeta
from pccm.source.core import Replace, Source, execute_modifiers
from pccm.utils import UniqueNamePool, get_qualname_of_type

PCCM_INLINE_MODULE_NAME = "inline_module"
PCCM_INLINE_FUNCTION_NAME = "inline_function"
PCCM_INLINE_INNER_FUNCTION_NAME = "inline_inner_function" # usually cuda global function

PCCM_INLINE_NAMESPACE = "inline_namespace"
PCCM_INLINE_CLASS_NAME = "InlineClass"


def gcs(*instances):
    if len(instances) == 0:
        return None
    classes = [inspect.getmro(type(x)) for x in instances]
    for x in classes[0]:
        if all(x in mro for mro in classes):
            return x


def get_base_type_string(obj):
    if isinstance(obj, int):
        return "int64_t", False
    elif isinstance(obj, float):
        return "float", False
    elif isinstance(obj, bool):
        return "bool", False
    elif isinstance(obj, str):
        return "std::string", False
    else:
        return get_qualname_of_type(type(obj)), True


class MultiTypeKindError(Exception):
    pass


class InlineBuilderPlugin:
    def handle_captured_type(self, name: str, code: FunctionCode,
                             obj: Any, user_arg: Optional[Any] = None) -> Optional[Tuple[str, str]]:
        raise NotImplementedError

    def type_conversion(self, obj: Any, user_arg: Optional[Any] = None):
        return obj

    def get_cpp_type(self, obj: Any, user_arg: Optional[Any] = None) -> str:
        raise NotImplementedError


def nested_type_analysis(obj,
                         plugin_dict: Dict[str, InlineBuilderPlugin],
                         iter_limit=10,
                         user_arg: Optional[Any] = None) -> Tuple[BaseType, BaseType]:
    if isinstance(obj, (list, set)):
        types_union: Set[CppType] = set()
        mapped_type = CppType([])
        for o in obj:
            if iter_limit < 0:
                break
            iter_limit -= 1
            type_s, type_mapped_s = nested_type_analysis(
                o, plugin_dict, iter_limit, user_arg)
            cpp_type = CppType([DeclSpec(type_s)])
            mapped_type = CppType([DeclSpec(type_mapped_s)])
            types_union.add(cpp_type)
        if len(types_union) != 1:
            raise MultiTypeKindError(
                "multiple type found, union type isn't supported.")
        if isinstance(obj, list):
            name = "vector"
        else:
            name = "unordered_set"
        mapped_type = BaseType(QualifiedId(["std", name]), [mapped_type])
        return BaseType(QualifiedId(["std", name]),
                        list(types_union)), mapped_type
    elif isinstance(obj, tuple):
        type_tuple: List[CppType] = []
        mapped_type_tuple: List[CppType] = []

        for o in obj:
            if iter_limit < 0:
                break
            iter_limit -= 1
            type_s, type_mapped_s = nested_type_analysis(
                o, plugin_dict, iter_limit, user_arg)

            mapped_type_tuple.append(CppType([DeclSpec(type_mapped_s)]))
            type_tuple.append(CppType([DeclSpec(type_s)]))
        mapped_type = BaseType(QualifiedId(["std", "tuple"]),
                               mapped_type_tuple)
        return BaseType(QualifiedId(["std", "tuple"]), type_tuple), mapped_type
    elif isinstance(obj, Mapping):
        key_types_union: Set[CppType] = set()
        val_type_union: Set[CppType] = set()
        key_mapped_type = CppType([])
        val_mapped_type = CppType([])

        for k, v in obj.items():
            if iter_limit < 0:
                break
            iter_limit -= 1
            key_type_s, key_m_s = nested_type_analysis(k, plugin_dict,
                                                       iter_limit, user_arg)
            val_type_s, val_m_s = nested_type_analysis(v, plugin_dict,
                                                       iter_limit, user_arg)
            key_mapped_type = CppType([DeclSpec(key_m_s)])
            val_mapped_type = CppType([DeclSpec(val_m_s)])

            key_types_union.add(CppType([DeclSpec(key_type_s)]))
            val_type_union.add(CppType([DeclSpec(val_type_s)]))
        if len(key_types_union) != 1:
            raise MultiTypeKindError(
                "multiple type found, union type isn't supported.")
        if len(val_type_union) != 1:
            raise MultiTypeKindError(
                "multiple type found, union type isn't supported.")
        res_mapped_type = BaseType(QualifiedId(["std", "unordered_map"]),
                                   [key_mapped_type, val_mapped_type])
        res_type = BaseType(
            QualifiedId(["std", "unordered_map"]),
            [list(key_types_union)[0],
             list(val_type_union)[0]])
        return res_type, res_mapped_type
    else:
        base_str, is_custom = get_base_type_string(obj)
        res_mapped = base_str
        if is_custom:
            res_mapped = plugin_dict[base_str].get_cpp_type(obj, user_arg)
        return BaseType(QualifiedId([base_str]),
                        []), BaseType(QualifiedId([res_mapped]), [])


class NameVisitor(ast.NodeVisitor):
    def __init__(self):
        self.contain_name = False
        self.names: List[str] = []

    def visit_Name(self, node: ast.Name):
        self.names.append(node.id)


def extract_names_from_expr(expr: str):
    tree = ast.parse(expr)
    vis = NameVisitor()
    vis.visit(tree)
    return vis.names


@dataclass
class CaptureStmt:
    name: str
    is_expr: bool
    range_pair: Tuple[int, int]
    expr_names: List[str]
    replaced_name: str
    replace_range_pairs: List[Tuple[int, int]]
    arg_name: str
    def __init__(self, name: str, is_expr: bool, range_pair: Tuple[int, int],
                 replace_range_pairs: List[Tuple[int, int]]) -> None:
        self.name = name
        self.is_expr = is_expr
        self.range_pair = range_pair
        self.replace_range_pairs = replace_range_pairs
        self.expr_names = []
        if is_expr:
            self.expr_names = extract_names_from_expr(name)
        self.replaced_name = name
        self.arg_name = name



def get_save_file(path: Path):
    pid = os.getpid()


def _nested_apply_plugin_transform(obj_type: BaseType, obj,
                                   plugins: Dict[str, InlineBuilderPlugin],
                                   user_arg: Optional[Any] = None):
    if obj_type.is_std_type():
        return obj
    qualname = obj_type.qualname
    if not qualname.startswith("std"):
        # custom type
        if qualname not in plugins:
            msg = f"can't find {qualname} in plugins, available: {plugins.keys()}"
            raise ValueError(msg)
        plugin = plugins[qualname]
        return plugin.type_conversion(obj, user_arg)

    elif qualname == "std::vector":
        res = []
        vt = obj_type.args[0].base_type
        for o in obj:
            o_res = _nested_apply_plugin_transform(vt, o, plugins, user_arg)
            res.append(o_res)
        return res
    elif qualname == "std::unordered_map":
        res = {}
        kt = obj_type.args[0].base_type
        vt = obj_type.args[1].base_type
        assert kt.is_simple_type(), "only support simple type for key"
        for (ok, ov) in obj.items():
            res[ok] = _nested_apply_plugin_transform(vt, ov, plugins, user_arg)
        return res
    elif qualname == "std::unordered_set":
        assert obj_type.is_std_type()
        return obj
    elif qualname == "std::tuple":
        res = []
        for o, ot in zip(obj, obj_type.args):
            ott = ot.base_type
            if ott.is_std_type():
                res.append(o)
            else:
                res.append(_nested_apply_plugin_transform(ott, o, plugins, user_arg))
        return tuple(res)
    else:
        raise NotImplementedError


def _expr_str_to_identifier(name: str):
    res = ""
    A = ord("A")
    Z = ord("Z")
    a = ord("a")
    z = ord("z")
    _0 = ord("0")
    _9 = ord("9")

    for s in name:
        c = ord(s)
        if (c >= A and c <= Z) or (c >= a and c <= z) or (c >= _0 and c <= _9):
            res += s
        elif s != " ":
            res += "_"
    return res


class NumpyPlugin(InlineBuilderPlugin):
    def handle_captured_type(self, name: str, code: FunctionCode,
                             obj: Any, user_arg: Optional[Any] = None) ->  Optional[Tuple[str, str]]:
        return

    def type_conversion(self, obj: Any, user_arg: Optional[Any] = None):
        return obj

    def get_cpp_type(self, obj: Any, user_arg: Optional[Any] = None) -> str:
        return "pybind11::array"


@dataclass
class ModuleMetaData:
    code: str
    deps: List[str]


_DEFAULT_PLUGINS: Dict[str, InlineBuilderPlugin] = {
    "numpy.ndarray": NumpyPlugin(),
}


class PyBind11(pccm.Class):
    def __init__(self):
        super().__init__()
        self.add_include("pybind11/stl.h")
        self.add_include("pybind11/pybind11.h")
        self.add_include("pybind11/numpy.h")


@pccm.pybind.bind_class_module_local
class InlineClass(Class):
    def __init__(self):
        super().__init__()
        self.add_include("vector", "unordered_map", "unordered_set", "tuple",
                         "string", "iostream", "fstream")
        self.add_dependency(PyBind11)

class _ModuleMeta:
    def __init__(self, func: Any, captures: List[CaptureStmt], capture_ctypes: List[BaseType], inner_code: str) -> None:
        self.func = func
        self.captures = captures
        self.capture_ctypes = capture_ctypes
        self.inner_code = inner_code

class InlineBuilder:
    """
    inliner.inline(...)    
    """
    def __init__(
            self,
            deps: List[Type[Class]],
            plugins: Optional[Dict[str, InlineBuilderPlugin]] = None,
            root: Optional[Path] = None,
            build_root: Optional[Path] = None,
            build_kwargs: Optional[Dict[str, Any]] = None,
            param_deps: Optional[List[pccm.ParameterizedClass]] = None,
            reload_when_code_change: bool = False) -> None:
        self.deps = deps
        if param_deps is None:
            param_deps = []
        self.param_deps = param_deps
        if plugins is None:
            self.plugins = _DEFAULT_PLUGINS
        else:
            self.plugins = plugins
        if build_kwargs is None:
            build_kwargs = {}
        self.build_kwargs = build_kwargs
        self.modules: Dict[Tuple[str, str, int], _ModuleMeta] = {}
        # self.cached_captures: Dict[Tuple[str, str, int],
        #                            List[CaptureStmt]] = {}
        # self.cached_capture_ctypes: Dict[Tuple[str, str, int],
        #                                  List[BaseType]] = {}

        self.lock = threading.Lock()

        self.dep_ids = [get_qualname_of_type(t) for t in self.deps]
        self.dep_ids.sort()
        self.used_names: Set[Tuple[str, str]] = set()
        self.root = root
        self.build_root = build_root
        self._reload_when_code_change = reload_when_code_change

    def _find_exist_module_name(self, code: str, code_hash: str, root: Path):
        cur_dep_ids = self.dep_ids
        for path in root.glob(f"_{code_hash}_*.json"):
            with path.open("r") as f:
                data = json.load(f)
            meta = ModuleMetaData(data["code"], data["deps"])
            if meta.code == code and meta.deps == cur_dep_ids:
                return path.stem
        return None

    def get_save_root(self, path: Path, root: Optional[Path] = None, build_root: Optional[Path] = None):
        if root is not None:
            relative_parts = path.resolve().parent.relative_to(root)
            import_parts = list(relative_parts.parts)
        else:
            import_parts = loader.try_capture_import_parts(path)
        if import_parts is None:
            raise NotImplementedError("you must use inline "
                                    "in a standard python project with "
                                    "pip installed.")
        if build_root is not None:
            res = build_root / "/".join(import_parts)
        else:
            res = PCCM_INLINE_LIBRARY_PATH / "/".join(import_parts)
        return res


    def handle_container_code(self, code_str: str, code: FunctionCode,
                    arg: Optional[Any]):
        code.raw(code_str)

    def create_inner_decl(self, code_str: str, container_fcode: FunctionCode, inner_fcode: FunctionCode,
                    arg: Optional[Any]) -> Optional[FunctionDecl]:
        
        return None 

    def build(self, pccm_cls: pccm.Class, mod_root: Path, name: str, timeout: float, user_arg: Optional[Any] = None):
        mod_root.mkdir(mode=0o755, parents=True, exist_ok=True)
        out_lib_path = mod_root / name
        build_dir = mod_root / name
        # out_lib_meta_path = mod_root / f"{prev_mod_name}.json"
        file_lock = mod_root / f"{name}.lock"
        # -1 is invalid for portalocker
        with portalocker.Lock(str(file_lock), timeout=timeout) as fh:
            mod = build_pybind([pccm_cls],
                                out_lib_path,
                                build_dir=build_dir,
                                **self.build_kwargs)

        return getattr(
            getattr(getattr(mod, PCCM_INLINE_NAMESPACE),
                    PCCM_INLINE_CLASS_NAME), PCCM_INLINE_FUNCTION_NAME)

    def run_func(self, func, *args, user_args: Optional[Any] = None):
        return func(*args)

    def get_base_class(self):
        return InlineClass()

    def inline(self,
               name: str,
               code: Union[str, FunctionCode],
               impl_file_suffix=".cc",
               additional_vars: Optional[Dict[str, Any]] = None,
               *,
               _frame_cnt: int = 1,
               user_arg: Optional[Any] = None,
               timeout: float=999999.0,
               disable_cache: bool = False):
        """use $var to capture python objects, use $(var.shape[0]) to capture anonymous expr.
        use different to handle different arg types.
        ~20-100us run overhead. 
        only support: 
        1. int/float/str and nested containers of int/float/str.
        2. custom type via plugins
        """
        if isinstance(code, FunctionCode):
            code_str = code.inspect_body()
        else:
            code_str = code
        if additional_vars is None:
            additional_vars = {}
        cur_frame = inspect.currentframe()
        assert cur_frame is not None
        frame = cur_frame
        while _frame_cnt > 0:
            frame = cur_frame.f_back
            assert frame is not None
            cur_frame = frame
            _frame_cnt -= 1
        # del frame
        local_vars = cur_frame.f_locals.copy()
        local_vars.update(additional_vars)
        code_path = cur_frame.f_code.co_filename
        lineno = cur_frame.f_lineno
        key = (code_path, name)
        unique_key = (code_path, name, lineno)
        if self._reload_when_code_change:
            unique_key = (code_path, name, -1)
        # del cur_frame
        with self.lock:
            if not disable_cache and not self._reload_when_code_change:
                exist = unique_key in self.modules
                if key in self.used_names and not exist:
                    raise ValueError("you use duplicate name in same file.", unique_key)
            else:
                exist = False
            if not exist:
                # 1. extract captured vars
                it = source_iter.CppSourceIterator(code_str)
                # hold ranges for further replace
                all_captures: List[CaptureStmt] = []
                unique_name: Dict[str, CaptureStmt] = {}
                for pose in it.get_symbol_poses("$"):
                    it.move(pose + 1)
                    next_round = it.next_round()
                    if next_round is not None:
                        cap_name = it.source[next_round[0] +
                                             1:next_round[1]].strip()
                        rep_range = (pose, next_round[1] + 1)
                        sym_range = (next_round[0] + 1, next_round[1])
                        is_expr = True
                    else:
                        iden = it.next_identifier()
                        assert iden is not None, "you can't use $ without a identifier."
                        rep_range = (pose, iden.end)
                        cap_name = iden.name.strip()
                        sym_range = (pose + 1, iden.end)
                        is_expr = False

                    if cap_name in unique_name:
                        cap = unique_name[cap_name]
                        cap.replace_range_pairs.append(rep_range)
                    else:
                        all_captures.append(
                            CaptureStmt(cap_name, is_expr, sym_range,
                                        [rep_range]))
                        unique_name[all_captures[-1].name] = all_captures[-1]
                # 2. find captures in prev frame

                container_fcode = FunctionCode()
                inner_fcode = FunctionCode()
                # 3. inference c++ types
                args = []
                name_pool = UniqueNamePool()
                replaces: List[Replace] = []
                capture_bts: List[BaseType] = []
                for cap in all_captures:
                    if not cap.is_expr:
                        if cap.name not in local_vars:
                            raise ValueError(
                                f"can't find your capture {cap.name} in prev frame."
                            )
                        obj = local_vars[cap.name]
                        arg_name = cap.name
                    else:
                        for cap_name in cap.expr_names:
                            if cap_name not in local_vars:
                                raise ValueError(
                                    f"can't find your capture {cap_name} in prev frame."
                                )
                        # eval expr in prev frame
                        obj = eval(cap.name, local_vars)
                    # apply non-anonymous vars (expr are anonymous vars)
                    try:
                        cpp_type, mapped_cpp_type = nested_type_analysis(
                            obj, self.plugins, user_arg=user_arg)
                    except:
                        print(f"ERROR: variable {cap.name} type analysis failed.")
                        raise
                    obj = _nested_apply_plugin_transform(
                        cpp_type, obj, self.plugins, user_arg)
                    if cap.is_expr:
                        cap.replaced_name = name_pool(
                            _expr_str_to_identifier(cap.replaced_name))
                    arg_name = cap.replaced_name
                    inner_cpp_type = str(mapped_cpp_type)
                    if not cap.is_expr:
                        if not cpp_type.is_std_type():
                            # custom type, must apply plugin
                            qualname = cpp_type.qualname
                            # here we only apply handle_captured_type on non-container custom type.
                            if qualname in self.plugins:
                                plugin = self.plugins[qualname]
                                res = plugin.handle_captured_type(
                                    cap.name, container_fcode, obj, user_arg)
                                if res is not None:
                                    new_arg_name, inner_cpp_type = res
                                    arg_name = new_arg_name

                    capture_bts.append(cpp_type)
                    args.append(obj)
                    cap.arg_name = arg_name
                    container_fcode.arg(arg_name, str(mapped_cpp_type))
                    inner_fcode.arg(cap.replaced_name, inner_cpp_type)
                    for rr in cap.replace_range_pairs:
                        replace = Replace(cap.replaced_name, *rr)
                        replaces.append(replace)
                for k, v in additional_vars.items():
                    _, mapped_cpp_type = nested_type_analysis(v, self.plugins, user_arg=user_arg)
                    container_fcode.arg(k, str(mapped_cpp_type))
                    args.append(v)

                inner_code_str = execute_modifiers(code_str, replaces)

                assert inner_code_str is not None
                if self._reload_when_code_change:
                    exist = unique_key in self.modules
                    if exist:
                        prev_inner_code = self.modules[unique_key].inner_code
                        if prev_inner_code == inner_code_str:
                            module_meta = self.modules[unique_key]
                            func = module_meta.func
                            all_captures = module_meta.captures
                            all_base_types = module_meta.capture_ctypes
                            args = []
                            for cap, bt in zip(all_captures, all_base_types):
                                if not cap.is_expr:
                                    if cap.name not in local_vars:
                                        raise ValueError(
                                            f"can't find your capture {cap.name} in prev frame.")
                                    obj = local_vars[cap.name]
                                else:
                                    for cap_name in cap.expr_names:
                                        if cap_name not in local_vars:
                                            raise ValueError(
                                                f"can't find your capture {cap_name} in prev frame."
                                            )
                                    # eval expr in prev frame
                                    obj = eval(cap.name, local_vars)
                                obj = _nested_apply_plugin_transform(bt, obj, self.plugins, user_arg)
                                args.append(obj)
                            for v in additional_vars.values():
                                args.append(v)
                            return self.run_func(func, *args, user_args=user_arg)
                if isinstance(code, FunctionCode):
                    container_fcode.ret(code.return_type)
                meta = self.handle_container_code(inner_code_str, container_fcode, user_arg)
                inner_decl = self.create_inner_decl(inner_code_str, container_fcode, inner_fcode, user_arg)

                # now we have complete code. we need to determine a history build dir and use it to build library if need.
                # here we must reserve build dir because we need to rebuild when dependency change.
                if meta is None:
                    meta = StaticMemberFunctionMeta(impl_file_suffix=impl_file_suffix)
                    meta.mw_metas.append(Pybind11MethodMeta())
                decl = FunctionDecl(meta, container_fcode)
                decl.meta.name = PCCM_INLINE_FUNCTION_NAME
                # container_fcode_str = decl.inspect_impl()
                # container_fcode_hash = hashlib.sha256(code_str.encode('utf-8')).hexdigest()

                pccm_class = self.get_base_class()
                pccm_class.class_name = PCCM_INLINE_CLASS_NAME
                pccm_class.add_func_decl(decl)

                for dep in decl.code._impl_only_deps:
                    pccm_class.add_impl_only_dependency_by_name(
                        decl.meta.name, dep.get_class_type())
                if isinstance(code, FunctionCode):
                    for dep in code._impl_only_deps:
                        pccm_class.add_impl_only_dependency_by_name(
                            decl.meta.name, dep.get_class_type())

                if inner_decl is not None:
                    inner_decl.meta.name = PCCM_INLINE_INNER_FUNCTION_NAME
                    pccm_class.add_func_decl(inner_decl)
                    for dep in inner_decl.code._impl_only_deps:
                        pccm_class.add_impl_only_dependency_by_name(
                            inner_decl.meta.name, dep.get_class_type())
                    if isinstance(code, FunctionCode):
                        for dep in code._impl_only_deps:
                            pccm_class.add_impl_only_dependency_by_name(
                                inner_decl.meta.name, dep.get_class_type())

                pccm_class.add_dependency(*self.deps)
                for pdep in self.param_deps:
                    pdep_ns = pdep.namespace
                    pdep_cls_name = pdep.get_user_provided_class_name()
                    assert pdep_ns is not None, "you must provide a namespace for param dep"
                    assert pdep_cls_name is not None, "you must provide a class_name for param dep"
                    # use user-defined class name as alias
                    pccm_class.add_param_class(pdep_cls_name, pdep, pdep_cls_name)
                pccm_class.namespace = PCCM_INLINE_NAMESPACE
                mod_root = self.get_save_root(Path(code_path), self.root, self.build_root)
                func = self.build(pccm_class, mod_root, name, timeout, user_arg)
                if not disable_cache:
                    module_meta = _ModuleMeta(func, all_captures, capture_bts, inner_code_str)
                    self.modules[unique_key] = module_meta
                    # self.cached_captures[unique_key] = all_captures
                    # self.cached_capture_ctypes[unique_key] = capture_bts
                    self.used_names.add(key)
                    # breakpoint()
                return self.run_func(func, *args, user_args=user_arg)

        # module already loaded. just run it after transform.
        module_meta = self.modules[unique_key]
        func = module_meta.func
        all_captures = module_meta.captures
        all_base_types = module_meta.capture_ctypes
        args = []
        for cap, bt in zip(all_captures, all_base_types):
            if not cap.is_expr:
                if cap.name not in local_vars:
                    raise ValueError(
                        f"can't find your capture {cap.name} in prev frame.")
                obj = local_vars[cap.name]
            else:
                for cap_name in cap.expr_names:
                    if cap_name not in local_vars:
                        raise ValueError(
                            f"can't find your capture {cap_name} in prev frame."
                        )
                # eval expr in prev frame
                obj = eval(cap.name, local_vars)
            obj = _nested_apply_plugin_transform(bt, obj, self.plugins, user_arg)
            args.append(obj)
        for v in additional_vars.values():
            args.append(v)
        return self.run_func(func, *args, user_args=user_arg)


def main():
    print(_expr_str_to_identifier("a.shape[0] + 5"))
    import ast

    import numpy as np
    tree = ast.parse("a.shape[0] + b + c")
    # print(ast.dump(tree))
    aa = np.array([1], dtype=np.float32)
    a = [aa, aa]
    # print(nested_type_analysis(aa.shape))
    b = InlineBuilder([])
    for i in range(10):

        code = FunctionCode().raw(f"""
        """)
        t = time.time()
        b.inline(
            "just_a_name", f"""
        // pybind::array
        float* ptr = reinterpret_cast<float*>($a[0].mutable_data());
        float* ptr2 = reinterpret_cast<float*>($a[1].mutable_data());

        ptr[0] += 1;
        ptr2[0] += 1;
        """)
        tt = time.time() - t
        print(tt)
        print(aa[0])




if __name__ == "__main__":
    main()



from typing import Any, Callable, Deque, Dict, List, Optional, Set, Tuple, Type, Union
from pccm.core.parsers import Attr

class Argument(object):
    def __init__(self,
                 name: str,
                 type: str,
                 default: Optional[str] = None,
                 array: Optional[str] = None,
                 pyanno: Optional[str] = None,
                 doc: Optional[str] = None,
                 attrs: Optional[List[Attr]] = None):
        self.name = name.strip()
        self.type_str = str(type).strip()  # type: str
        self.default = default
        self.array = array
        self.pyanno = pyanno
        self.doc = doc
        if attrs is None:
            self.attrs: List[Attr] = []
        else:
            self.attrs: List[Attr] = attrs
        if pyanno is not None:
            self.pyanno = pyanno.strip()
            assert len(pyanno) != 0

class TemplateTypeArgument(object):
    def __init__(self,
                 name: str,
                 default: Optional[str] = None,
                 template: str = "",
                 packed: bool = False):
        self.name = name
        self.template = template
        self.packed = packed
        self.default = default

    def to_string(self) -> str:
        pack = ""
        if self.packed:
            if self.default:
                raise ValueError("packed arg can't have default value")
            pack = "..."
        if self.template:
            res = "{} typename{} {}".format(self.template, pack, self.name)
        else:
            res = "typename{} {}".format(pack, self.name)
        if self.default:
            res += " = {}".format(self.default)
        return res

class TemplateNonTypeArgument(object):
    def __init__(self,
                 name: str,
                 type: str,
                 default: Optional[str] = None,
                 packed: bool = False):
        self.name = name
        self.type = type
        self.packed = packed
        self.default = default

    def to_string(self) -> str:
        pack = ""
        if self.packed:
            if self.default:
                raise ValueError("packed arg can't have default value")
            pack = "..."
        res = "{}{} {}".format(self.type, pack, self.name)
        if self.default:
            res += " = {}".format(self.default)
        return res



import pccm
from pccm.libs import stl


class Test1(pccm.Class):
    def __init__(self):
        super().__init__()
        self.add_dependency(stl.STL)

    @pccm.member_function(const=True, virtual=True)
    def add(self):
        return pccm.FunctionCode(
            """
        if (a > 0){
            return b;
        }
        return a + b;
        """, [pccm.Argument("a", "int"),
              pccm.Argument("b", "int")], "int")


class PTest1(pccm.ParameterizedClass):
    def __init__(self, const=5, type="float"):
        super().__init__()
        self.const = const
        self.type = type

    @pccm.member_function
    def add_template(self):
        argus = [pccm.Argument("a", self.type), pccm.Argument("b", self.type)]
        return pccm.FunctionCode("""
        return a + b + {};
        """.format(self.const),
                                 argus,
                                 return_type=self.type)

    @pccm.member_function
    def add_func_template(self):
        code = pccm.FunctionCode("""
        return a + b + {};
        """.format(self.const))
        code.targ("T")
        code.arg("a,b", "T")
        code.ret(self.type)
        return code


class Test2(pccm.Class):
    def __init__(self):
        super().__init__()
        self.add_dependency(Test1)
        self.build_meta.add_private_cflags("g++", "-fvisibility=hidden")
        self.build_meta.add_public_cflags("g++", "-Weffc++")

    @pccm.member_function
    def add(self):
        return pccm.FunctionCode(
            """
        auto tst1 = Test1();
        return tst1.add(a, b);
        """, [pccm.Argument("a", "int"),
              pccm.Argument("b", "int")], "int")



from typing import Any, Callable, Dict, Hashable, Optional


class TypeHandler(object):
    @staticmethod
    def register_handler(global_dict, *class_types):
        def wrap_func(handler):
            for cls_type in class_types:
                global_dict[cls_type] = handler

            def new_handler(obj):
                return handler(obj)

            return handler

        return wrap_func

    @staticmethod
    def get_handler(global_dict, obj) -> Optional[Callable]:
        # 1. fast check
        obj_type = type(obj)
        if obj_type in global_dict:
            return global_dict[obj_type]
        else:
            # 2. iterate and isinstance
            for class_type, handler in global_dict.items():
                if isinstance(obj, class_type):
                    return handler
        return None

    @staticmethod
    def get_type_handler(global_dict, obj_type) -> Optional[Callable]:
        # 1. fast check
        if obj_type in global_dict:
            return global_dict[obj_type]
        else:
            # 2. iterate and isinstance
            for class_type, handler in global_dict.items():
                if issubclass(obj_type, class_type):
                    return handler
        return None





import bisect
import dataclasses
from typing import List, Tuple
from pathlib import Path 


@dataclasses.dataclass
class Replace:
    content: str 
    start: int 
    end: int # [start:end], [], not[)

def interval_remain(interval_sorted: List[Tuple[int, int]], start: int, end: int):
    before = start 
    res: List[Tuple[int, int]] = []
    for x, y in interval_sorted:
        # if x != before:
        res.append((before, x))
        before = y 
    res.append((before, end))
    return res 


class Source:
    source: str 
    length: int 
    line_offsets: List[Tuple[int, int]]
    lines: List[str]
    def __init__(self, source: str) -> None:
        self.source = source 
        self.length = len(source)
        self.line_offsets = []
        self.lines = []
        start = 0
        for i in range(self.length):
            val = source[i]
            if val == "\n":
                self.line_offsets.append((start, i))
                self.lines.append(source[start:i])
                start = i + 1
        self.line_offsets.append((start, self.length))
        self.line_offset_starts = [x[0] for x in self.line_offsets]

        self.lines.append(source[start:self.length])

    @property
    def num_lines(self):
        return len(self.lines)

    def lineno_col_to_offset(self, lineno: int, column: int):
        lineno -= 1
        column -= 1
        if lineno < 0 or lineno >= self.num_lines:
            return -1 
        start, end = self.line_offsets[lineno]
        if column < 0:
            column += end - start + 1
        if column >= end:
            return -1 
        return start + column 

    def offset_to_lineno_col(self, offset: int):
        if offset < 0 or offset > self.length:
            return (-1, -1)
        pos = bisect.bisect_left(self.line_offset_starts, offset)
        if pos == len(self.line_offsets):
            return (self.num_lines, offset - self.line_offsets[-1][0] + 1)
        
        start, end = self.line_offsets[pos - 1]
        return (pos, offset - start + 1)

    def get_content(self, lineno: int, column: int, end_lineno: int, end_column: int):
        start = self.lineno_col_to_offset(lineno, column)
        end = self.lineno_col_to_offset(end_lineno, end_column)
        return self.source[start:end + 1]

    def get_content_by_offset(self, start: int, end: int):
        return self.source[start:end + 1]


def execute_modifiers(source: str, modifiers: List[Replace]):
    if not modifiers:
        return source 

    modifiers.sort(key=lambda x: x.start)
    before = -1
    intervals: List[Tuple[int, int]] = []
    for m in modifiers:
        if m.start < before:
            return None 
        if m.end < m.start:
            return None 
        before = m.end
        intervals.append((m.start, m.end))
    if before > len(source):
        return None 

    remain_intervals = interval_remain(intervals, 0, len(source))
    res = ""
    index = 0
    for x, y in remain_intervals:
        res += source[x:y]
        if index < len(modifiers):
            res += modifiers[index].content
        index += 1
    return res 


def main():
    with (Path(__file__)).open("r") as f:
        content = f.read()

    s = Source(content)

    print(s.get_content(83, 1, 83, -1))

if __name__ == "__main__":
    main()


# Copyright 2022 Yan Yan
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Type, Union

import numpy as np
from pccm import FunctionCode
from pccm.builder.inliner import InlineBuilder, InlineBuilderPlugin
from pccm.utils import get_qualname_of_type
import numpy.typing as npt

class ArrayNumpy:
    def __init__(self, arr: np.ndarray, dtype: Optional[npt.DTypeLike] = None) -> None:
        assert arr.reshape(-1).shape[0] < 50, "array too large, must < 50"
        self.arr = arr
        self.dtype = np.dtype(dtype)
        

_NPDTYPE_TO_LIST_TYPE_STR: Dict[np.dtype, str] = {
    np.dtype(np.float16): "float",
    np.dtype(np.float32): "float",
    np.dtype(np.float64): "float",
    np.dtype(np.int8): "int64_t",
    np.dtype(np.int16): "int64_t",
    np.dtype(np.int32): "int64_t",
    np.dtype(np.int64): "int64_t",
    np.dtype(np.uint8): "int64_t",
    np.dtype(np.uint16): "int64_t",
    np.dtype(np.uint32): "int64_t",
    np.dtype(np.uint64): "int64_t",
    np.dtype(np.bool_): "bool",
}

class ArrayNumpyPlugin(InlineBuilderPlugin):
    """used to capture numpy to std::array.
    """
    QualifiedName = get_qualname_of_type(ArrayNumpy)
    def handle_captured_type(self, name: str, code: FunctionCode, obj: Any,
                                   user_arg: Optional[Any] = None) -> Optional[str]:
        return

    def type_conversion(self, obj: ArrayNumpy,
                                   user_arg: Optional[Any] = None):
        return obj.arr.tolist()

    def get_cpp_type(self, obj: ArrayNumpy,
                                   user_arg: Optional[Any] = None) -> str:
        ndim = obj.arr.ndim
        dtype = obj.dtype 
        if dtype is None:
            dtype = obj.arr.dtype 
        cpp_type = _NPDTYPE_TO_LIST_TYPE_STR[dtype]
        array_type = ""
        array_type += "std::array<" * ndim
        array_type += f"{cpp_type}, "
        shape_rev = obj.arr.shape[::-1]
        array_type += ">, ".join(map(str, shape_rev))
        # std::array<std::array<std::array<float, 3>, 4>, 2>
        return array_type + ">"

def main():
    import numpy as np 
    aa = np.zeros([2, 3, 4], dtype=np.float32)
    aa_array = ArrayNumpy(aa)
    b = InlineBuilder([], {ArrayNumpyPlugin.QualifiedName: ArrayNumpyPlugin()})
    for i in range(10):
        b.inline("just_a_name", f"""
        std::cout << $aa_array[0][0][0] << std::endl;
        """)

if __name__ == "__main__":
    main()



from typing import Literal, Optional, Union


class EnumClassValue(object):
    def __init__(self, value: int):
        self.__value = value

    def __int__(self):
        return self.__value

    def __index__(self) -> int:
        return self.__value

    @property
    def value(self):
        return self.__value

    @property
    def name(self) -> str:
        ...


class EnumValue(EnumClassValue):
    def __eq__(self, other: Optional["EnumValue"]) -> bool:
        ...

    def __ne__(self, other: Optional["EnumValue"]) -> bool:
        ...

    def __lt__(self, other: Union["EnumValue", int]) -> bool:
        ...

    def __gt__(self, other: Union["EnumValue", int]) -> bool:
        ...

    def __le__(self, other: Union["EnumValue", int]) -> bool:
        ...

    def __ge__(self, other: Union["EnumValue", int]) -> bool:
        ...

    def __and__(self, other: Union["EnumValue", int]) -> int:
        ...

    def __rand__(self, other: Union["EnumValue", int]) -> int:
        ...

    def __or__(self, other: Union["EnumValue", int]) -> int:
        ...

    def __ror__(self, other: Union["EnumValue", int]) -> int:
        ...

    def __xor__(self, other: Union["EnumValue", int]) -> int:
        ...

    def __rxor__(self, other: Union["EnumValue", int]) -> int:
        ...

    def __invert__(self) -> int:
        ...



import importlib.util
import os
import sys
from typing import Type, TypeVar
import re
from pathlib import Path

_TYPE_STRING_PATTERN = re.compile(r"<class '(.*)'>")


def project_is_editable(proj_name: str):
    """Is distribution an editable install?"""
    spec = importlib.util.find_spec(proj_name)
    if spec is None or spec.origin is None:
        raise ModuleNotFoundError(
            f"{proj_name} not found. you need to install it by pip.")
    git_ignore_check = (Path(spec.origin).parent.parent /
                        ".gitignore").exists()
    setuppy_check = (Path(spec.origin).parent.parent / "setup.py").exists()
    # if we find gitignore or setup.py in a package, this package is editable
    # installed for all of my python project.
    # for project with custom path (e.g. src/pkg_name), this check doesn't work.
    if setuppy_check or git_ignore_check:
        return True
    # following check doens't work in some environment.
    for path_item in sys.path:
        egg_link = os.path.join(path_item, proj_name + '.egg-link')
        if os.path.isfile(egg_link):
            return True
    return False


def project_is_installed(proj_name: str):
    """Is distribution an editable install?"""
    spec = importlib.util.find_spec(proj_name)
    if spec is None or spec.origin is None:
        return False
    return True


def _make_unique_name(unique_set, name, max_count=10000):
    if name not in unique_set:
        unique_set.add(name)
        return name
    for i in range(max_count):
        new_name = name + "_{}".format(i)
        if new_name not in unique_set:
            unique_set.add(new_name)
            return new_name
    raise ValueError("max count reached")


def get_qualname_of_type(klass: Type) -> str:
    module = klass.__module__
    if module == 'builtins':
        return klass.__qualname__  # avoid outputs like 'builtins.str'
    return module + '.' + klass.__qualname__


class UniqueNamePool:
    def __init__(self, max_count=10000):
        self.max_count = max_count
        self.unique_set = set()

    def __call__(self, name):
        return _make_unique_name(self.unique_set, name, self.max_count)

    def __contains__(self, key):
        return key in self.unique_set



import inspect
from typing import Any, Optional, Type 

def get_members(obj: Any, no_parent: bool = True, python_inherit: Optional[Type] = None):
    """this function return member functions that keep def order.
    """
    this_cls = type(obj)
    if not no_parent:
        assert python_inherit is None, "not implemented"
        res = inspect.getmembers(this_cls, inspect.isfunction)
        # inspect.getsourcelines need to read file, so .__code__.co_firstlineno
        # is greatly faster than it.
        # res.sort(key=lambda x: inspect.getsourcelines(x[1])[1])
        res.sort(key=lambda x: x[1].__code__.co_firstlineno)
        return res
    parents = inspect.getmro(this_cls)[1:]
    parents_methods = set()
    for parent in parents:
        if python_inherit is not None:
            if parent is python_inherit:
                # add inherited python members
                continue
        members = inspect.getmembers(parent, predicate=inspect.isfunction)
        parents_methods.update(members)

    child_methods = set(
        inspect.getmembers(this_cls, predicate=inspect.isfunction))
    child_only_methods = child_methods - parents_methods
    res = list(child_only_methods)
    # res.sort(key=lambda x: inspect.getsourcelines(x[1])[1])
    res.sort(key=lambda x: x[1].__code__.co_firstlineno)
    return res

